# -*- coding: UTF-8 -*-
# @Time    : 18-4-27
# @File    : demo_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
from antgo.utils import logger
import os
import shutil
import json
import numpy as np
from antgo.utils.fs import *
from antgo.utils.encode import *
from PIL import Image
import imageio
import uuid
import signal
from ..utils.serialize import loads,dumps
from antgo.crowdsource.utils import *
import functools


class BaseHandler(tornado.web.RequestHandler):
  @property
  def demo_name(self):
    return self.settings.get('demo_name', '-')

  @property
  def demo_type(self):
    return self.settings.get('demo_type', '-')

  @property
  def demo_description(self):
    return self.settings.get('demo_description',{})

  @property
  def demo_dump(self):
    return self.settings.get('demo_dump', '')

  @property
  def demo_support_user_upload(self):
    return self.settings.get('support_user_upload', True)

  @property
  def demo_support_user_input(self):
    return self.settings.get('support_user_input', True)

  @property
  def demo_support_user_interaction(self):
    return self.settings.get('support_user_interaction', False)

  @property
  def demo_support_user_comment(self):
    return self.settings.get('support_user_comment', False)

  @property
  def html_template(self):
    return self.settings['html_template']
  
  @property
  def db(self):
    return self.settings['db']
  
  @property
  def port(self):
    return self.settings['port']

  @property
  def demo_constraint(self):
    constraint = self.settings.get('support_user_constraint', None)
    if constraint is None:
      constraint = {}

    return constraint

  @property
  def dataset_queue(self):
    return self.settings['dataset_queue']

  @property
  def recorder_queue(self):
    return self.settings['recorder_queue']

  def _transfer(self, data_key, data_value, data_type, demo_response):
    if data_type == 'IMAGE':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output'))
        demo_response['DATA'].update({data_key: {'DATA': '/static/output/%s' % data.split('/')[-1], 'TYPE': 'IMAGE'}})
    elif data_type == 'VIDEO':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output'))
        demo_response['DATA'].update({data_key: {'DATA': '/static/output/%s' % data.split('/')[-1], 'TYPE': 'VIDEO'}})
    elif data_type == 'AUDIO':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output'))
        demo_response['DATA'].update({data_key: {'DATA': '/static/output/%s' % data.split('/')[-1], 'TYPE': 'AUDIO'}})
    elif data_type == 'FILE':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output'))
        demo_response['DATA'].update({data_key: {'DATA': '/static/output/%s' % data.split('/')[-1], 'TYPE': 'FILE'}})
    else:
      # string
      demo_response['DATA'].update({data_key: {'DATA': str(data_value), 'TYPE': 'STRING'}})

    return demo_response

  def preprocess_model_server(self, data, data_type):
    if data_type == 'URL' or data_type == 'PATH':
      data_path = os.path.normpath(data)
      data_name = data_path.split('/')[-1]
      ext_name = data_path.split('/')[-1].split('.')[-1].lower()

      if ext_name in ['jpg', 'jpeg', 'png', 'bmp', 'gif']:
        return data_path, data_name, 'IMAGE'
      elif ext_name in ['mp4', 'avi']:
        return data_path, data_name, 'VIDEO'
      elif ext_name in ['txt']:
        return data_path, data_name, 'FILE'
      else:
        # TODO: support video and sound
        logger.warn('dont support file type %s' % ext_name)
        return None, None, None
    else:
      return data, str(uuid.uuid4()), 'STRING'

  def postprocess_model_server(self, uuid_flag, demo_result):
    demo_predict = None
    demo_predict_additional = []
    if type(demo_result) == list or type(demo_result) == tuple:
      demo_predict, demo_predict_additional = demo_result[0:2]
    else:
      demo_predict = demo_result

    # build output folder (static/output)
    if not os.path.exists(os.path.join(self.demo_dump, 'static', 'output')):
      os.makedirs(os.path.join(self.demo_dump, 'static', 'output'))

    demo_response = {'DATA': {}}
    if demo_predict['DATA'] is not None and demo_predict['TYPE'] is not None:
      demo_response = self._transfer('RESULT', demo_predict['DATA'], demo_predict['TYPE'], demo_response)
    else:
      demo_response = {'DATA': {'RESULT': {'DATA':'', 'TYPE': 'STRING'}}}

    for data in demo_predict_additional:
      data_type = data['TYPE']
      data_value = None
      data_key = None
      for k,v in data.items():
        if k != 'TYPE':
          data_key = k
          data_value = v
          break

      if data_key is not None and data_value is not None:
        demo_response = self._transfer(data_key, data_value, data_type, demo_response)

    return demo_response

    
class IndexHandler(BaseHandler):
  def get(self):
    image_history_data = []
    history_folder = os.path.join(self.demo_dump, 'static', 'input')
    if os.path.exists(history_folder):
      for f in os.listdir(history_folder):
        ext_name = f.split('.')[-1]
        if ext_name.lower() in ['jpg', 'jpeg', 'png', 'gif']:
          image_history_data.append('/static/input/%s'%f)
    
    input_filter = ''

    if 'file_type' in self.demo_constraint:
      for support_format in self.demo_constraint['file_type']:
        if support_format.lower() in ['jpg', 'jpeg', 'png', 'gif']:
          input_filter += 'image/%s,'%support_format.lower()
        elif support_format.lower() in ['mp4']:
          input_filter += 'video/%s,'%support_format.lower()
        elif support_format.lower() in ['mp3', 'wav']:
          input_filter += 'audio/%s,'%support_format.lower()
        else:
          input_filter += '%s,'%support_format.lower()

    if len(input_filter) > 0:
      input_filter = input_filter[0:-1]
    else:
      input_filter = '*'

    self.render(self.html_template, demo={'name': self.demo_name,
                                          'type': self.demo_type,
                                          'description': self.demo_description,
                                          'upload': self.demo_support_user_upload,
                                          'upload_accept': input_filter,
                                          'input': self.demo_support_user_input,
                                          'image_history': image_history_data})


class ClientQueryHandler(BaseHandler):
  def _file_download(self,file_name, response):
    download_path = os.path.join(self.demo_dump, 'static', 'input', file_name)
    with open(download_path, "wb") as f:
        f.write(response.body)

  @gen.coroutine
  def post(self):
    # 0.step check support status
    if not self.demo_support_user_upload:
      self.set_status(500)
      self.write(json.dumps({'code': 'InvalidSupport', 'message': 'demo dont support upload'}))
      self.finish()
      return

    upload_path = os.path.join(self.demo_dump, 'static', 'input')
    if not os.path.exists(upload_path):
      os.makedirs(upload_path)

    # 1.step parse query data
    model_datas = []
    model_data_names = []
    model_data_types = []
    request_param = None

    data = self.get_argument('DATA', None)
    data_type = self.get_argument('DATA_TYPE', None)
    if data is not None and data_type is not None:
      if data_type == 'URL':
        # download data from url
        http_client = httpclient.AsyncHTTPClient()
        file_name = os.path.normpath(data).split('/')[-1]
        if file_name == '':
          file_name = '%s'%str(uuid.uuid4())
        file_download_func = functools.partial(self._file_download, file_name)
        yield http_client.fetch(data, callback=file_download_func)

        file_path = os.path.join(self.demo_dump, 'static', 'input', file_name)
        if 'file_size' in self.demo_constraint:
          max_file_size = self.demo_constraint['file_size']
          fsize = os.path.getsize(file_path) / float(1024 * 1024)
          if round(fsize,2) > max_file_size:
            self.set_status(400)
            self.write(json.dumps({'code': 'InvalidImageSize', 'message': 'The input file size is too large (>%f MB)'%float(max_file_size)}))
            self.finish()
            return

        # 2.step check file format
        if 'file_type' in self.demo_constraint:
          is_ok, file_path = check_file_types(file_path, self.demo_constraint['file_type'])
          if not is_ok:
            self.set_status(400)
            self.write(json.dumps({'code': 'InvalidImageFormat', 'message': 'The input file is not in a valid image format that the service can support'}))
            self.finish()
            return

        # DATA_TYPE, DATA, TASK_NAME, TASK_TYPE
        model_data, model_data_name, model_data_type = self.preprocess_model_server(file_path, data_type)
        if model_data is None:
          self.set_status(500)
          self.write(json.dumps({'code': 'InvalidIO', 'message': 'couldnt parse data'}))
          self.finish()
          return

        model_datas.append(model_data)
        model_data_names.append(model_data_name)
        model_data_types.append(model_data_type)
    else:
      if len(self.request.files) == 0:
        self.set_status(400)
        self.write(json.dumps({'code': 'InvalidUploadFile', 'message': 'The input file is not uploaded correctly'}))
        self.finish()
        return

      file_paths = []
      file_names = []
      for _, meta in self.request.files.items():
        _file_name = '%s_%s' % (str(uuid.uuid4()), meta[0]['filename'])
        _file_path = os.path.join(upload_path, _file_name)

        with open(_file_path, 'wb') as fp:
          fp.write(meta[0]['body'])

        file_paths.append(_file_path)
        file_names.append(_file_name)

      if len(file_paths) == 0 or len(file_names) == 0:
        self.set_status(400)
        self.write(json.dumps({'code': 'InvalidUploadFile', 'message': 'The input file is not uploaded correctly'}))
        self.finish()
        return

      # 2.step parse query data
      for file_path in file_paths:
        # check file basic infomation
        # 1.step check file size
        if 'file_size' in self.demo_constraint:
          max_file_size = self.demo_constraint['file_size']
          fsize = os.path.getsize(file_path) / float(1024 * 1024)
          if round(fsize, 2) > max_file_size:
            self.set_status(400)
            self.write(json.dumps({'code': 'InvalidImageSize',
                                   'message': 'The input file size is too large (>%f MB)' % float(max_file_size)}))
            self.finish()
            return

        # 2.step check file format
        if 'file_type' in self.demo_constraint:
          is_ok, file_path = check_file_types(file_path, self.demo_constraint['file_type'])
          if not is_ok:
            self.set_status(400)
            self.write(json.dumps({'code': 'InvalidImageFormat',
                                   'message': 'The input file is not in a valid image format that the service can support'}))
            self.finish()
            return

        try:
          model_data, model_data_name, model_data_type = self.preprocess_model_server(file_path, 'PATH')
        except:
          self.set_status(400)
          self.write(json.dumps({'code': 'InvalidDetails', 'message': 'The input file parse error'}))
          self.finish()
          return

        if model_data is None:
          self.set_status(400)
          self.write(json.dumps({'code': 'InvalidDetails', 'message': 'The input file parse error'}))
          self.finish()
          return

        model_datas.append(model_data)
        model_data_names.append(model_data_name)
        model_data_types.append(model_data_type)

      request_param = self.get_argument('params', None)
      if request_param is not None:
        request_param = json.loads(request_param)

    # push to backend
    model_input = (model_datas, model_data_types, request_param) if len(model_datas) > 1 else (model_datas[0], model_data_types[0], request_param)
    self.dataset_queue.put(model_input)

    # wating to get result from backend
    result = self.recorder_queue.get()
    _, demo_result = result

    # data type: 'FILE', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'
    # data:       PATH,  '',        PATH,    PATH,    PATH
    # 5.step post process and render
    demo_response = self.postprocess_model_server(model_data_names[0], demo_result)

    if len(model_data_names) == 1:
      demo_response['INPUT_TYPE'] = model_data_types[0]
      if model_data_types[0] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
        demo_response['INPUT'] = '/static/input/%s' % model_data_names[0]
      else:
        demo_response['INPUT'] = model_datas[0]
    else:
      demo_response['INPUT_TYPE'] = []
      demo_response['INPUT'] = []
      for index in range(len(model_data_names)):
        demo_response['INPUT_TYPE'].append(model_data_types[index])
        if model_data_types[index] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
          demo_response['INPUT'].append('/static/input/%s' % model_data_names[index])
        else:
          demo_response['INPUT'].append(model_datas[index])

    self.write(json.dumps(demo_response))
    self.finish()


class ClientFileUploadAndProcessHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 0.step check support status
    if not self.demo_support_user_upload:
      self.set_status(500)
      self.write(json.dumps({'code': 'InvalidSupport', 'message': 'demo dont support upload'}))
      self.finish()
      return

    # 1.step receive client upload file
    upload_path = os.path.join(self.demo_dump, 'static', 'input')
    if not os.path.exists(upload_path):
      os.makedirs(upload_path)

    file_metas = self.request.files.get('file', None)
    if not file_metas:
      self.set_status(400)
      self.write(json.dumps({'code': 'InvalidUploadFile', 'message': 'The input file is not uploaded correctly'}))
      self.finish()
      return

    file_paths = []
    file_names = []
    for meta in file_metas:
      _file_name = '%s_%s'%(str(uuid.uuid4()), meta['filename'])
      _file_path = os.path.join(upload_path, _file_name)

      with open(_file_path, 'wb') as fp:
        fp.write(meta['body'])

      file_paths.append(_file_path)
      file_names.append(_file_name)

    if len(file_paths) == 0 or len(file_names) == 0:
      self.set_status(400)
      self.write(json.dumps({'code': 'InvalidUploadFile', 'message': 'The input file is not uploaded correctly'}))
      self.finish()
      return

    # 2.step parse query data
    model_datas = []
    model_data_names = []
    model_data_types = []
    request_param = None

    for file_path in file_paths:
      # check file basic infomation
      # 1.step check file size
      if 'file_size' in self.demo_constraint:
        max_file_size = self.demo_constraint['file_size']
        fsize = os.path.getsize(file_path) / float(1024 * 1024)
        if round(fsize,2) > max_file_size:
          self.set_status(400)
          self.write(json.dumps({'code': 'InvalidImageSize', 'message': 'The input file size is too large (>%f MB)'%float(max_file_size)}))
          self.finish()
          return

      # 2.step check file format
      if 'file_type' in self.demo_constraint:
        is_ok, file_path = check_file_types(file_path, self.demo_constraint['file_type'])
        if not is_ok:
          self.set_status(400)
          self.write(json.dumps({'code': 'InvalidImageFormat', 'message': 'The input file is not in a valid image format that the service can support'}))
          self.finish()
          return
      try:
        model_data, model_data_name, model_data_type = self.preprocess_model_server(file_path, 'PATH')
      except:
        self.set_status(400)
        self.write(json.dumps({'code': 'InvalidDetails', 'message': 'The input file parse error'}))
        self.finish()
        return
      if model_data is None:
        self.set_status(400)
        self.write(json.dumps({'code': 'InvalidDetails', 'message': 'The input file parse error'}))
        self.finish()
        return

      model_datas.append(model_data)
      model_data_names.append(model_data_name)
      model_data_types.append(model_data_type)

    # 3.step preprocess data, then submit to model
    model_input = (model_datas,model_data_types,request_param) if len(model_datas) > 1 else (model_datas[0], model_data_types[0], request_param)

    self.dataset_queue.put(model_input)
    result = self.recorder_queue.get()
    _, demo_result = result

    # data type: 'FILE', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'
    # data:       PATH,  '',        PATH,    PATH,    PATH
    # 5.step post process and render
    demo_response = self.postprocess_model_server(model_data_names[0], demo_result)

    if len(model_data_names) == 1:
      demo_response['INPUT_TYPE'] = model_data_types[0]
      if model_data_types[0] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
        demo_response['INPUT'] = '/static/input/%s' % model_data_names[0]
      else:
        demo_response['INPUT'] = model_datas[0]
    else:
      demo_response['INPUT_TYPE'] = []
      demo_response['INPUT'] = []
      for index in range(len(model_data_names)):
        demo_response['INPUT_TYPE'].append(model_data_types[index])
        if model_data_types[index] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
          demo_response['INPUT'].append('/static/input/%s' % model_data_names[index])
        else:
          demo_response['INPUT'].append(model_datas[index])

    self.write(json.dumps(demo_response))
    self.finish()


class ClientCommentHandler(BaseHandler):
  def post(self):
    self.set_status(201)
    self.finish()


class PrefixRedirectHandler(BaseHandler):
  def get(self):
    static_pi = self.request.uri.find('static')
    path = self.request.uri[static_pi:]
    self.redirect('http://127.0.0.1:%d/%s'%(self.port, path), permanent=False)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def demo_server_start(demo_name,
                      demo_type,
                      demo_config,
                      parent_id,
                      dataset_queue,
                      recorder_queue):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)
  
  try:
    # 0.step define http server port
    define('port', default=demo_config['port'], help='run on port')

    # 1.step prepare static resource
    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    demo_static_dir = os.path.join(demo_config['dump_dir'], 'static')
    if not os.path.exists(demo_static_dir):
      os.makedirs(demo_static_dir)
  
    for static_file in os.listdir(os.path.join(static_folder, 'resource', 'static')):
      if static_file[0] == '.':
        continue
    
      shutil.copy(os.path.join(static_folder, 'resource', 'static', static_file), demo_static_dir)
  
    # 2.step prepare html template
    demo_tempate_dir = os.path.join(demo_config['dump_dir'], 'templates')
  
    if not os.path.exists(demo_tempate_dir):
      os.makedirs(demo_tempate_dir)

    html_template = demo_config['html_template']
    if html_template is None:
      html_template = 'demo.html'
  
      if not os.path.exists(os.path.join(demo_tempate_dir, html_template)):
        assert(os.path.exists(os.path.join(static_folder, 'resource', 'templates',html_template)))
        shutil.copy(os.path.join(static_folder, 'resource', 'templates',html_template), demo_tempate_dir)
    else:
      shutil.copy(html_template, demo_tempate_dir)

    tornado.options.parse_command_line()
    settings = {
      'template_path': demo_tempate_dir,
      'static_path': demo_static_dir,
      'html_template': html_template,
      'port': demo_config['port'],
      'demo_dump': demo_config['dump_dir'],
      'demo_name': demo_name,
      'demo_type': demo_type,
      'demo_description': demo_config['description_config'],
      'support_user_upload': demo_config['interaction']['support_user_upload'],
      'support_user_input': demo_config['interaction']['support_user_input'],
      'support_user_interaction': demo_config['interaction']['support_user_interaction'],
      'support_user_constraint': demo_config['interaction']['support_user_constraint'],
      'dataset_queue': dataset_queue,
      'recorder_queue': recorder_queue
    }
    app = tornado.web.Application(handlers=[(r"/", IndexHandler),
                                            (r"/demo", IndexHandler),
                                            (r"/api/query/", ClientQueryHandler),
                                            (r"/api/comment/", ClientCommentHandler),
                                            (r"/submit/", ClientFileUploadAndProcessHandler),
                                            (r"/.*/static/.*", PrefixRedirectHandler)],
      **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)
  
    logger.info('demo is providing server on port %d'%demo_config['port'])
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    os.kill(parent_id, signal.SIGKILL)
    



# -*- coding: UTF-8 -*-
# @Time : 17/03/2018
# @File : utils.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import imghdr
import os
import shutil

def url_path_join(*pieces):
  """Join components of url into a relative url

  Use to prevent double slash when joining subpath. This will leave the
  initial and final / in place

  Copied from notebook.utils.url_path_join
  """
  initial = pieces[0].startswith('/')
  final = pieces[-1].endswith('/')
  stripped = [s.strip('/') for s in pieces]
  result = '/'.join(s for s in stripped if s)

  if initial:
    result = '/' + result
  if final:
    result = result + '/'
  if result == '//':
    result = '/'

  return result


def check_file_types(file_path, check_types=['jpg']):
  if check_types[0] in ['jpg', 'jpeg', 'png', 'bmp', 'gif']:
    # IMAGE
    file_type = imghdr.what(file_path)
    file_path = os.path.normpath(file_path)
    file_name = file_path.split('/')[-1]
    if file_type != file_name.split('.')[-1]:
      file_name = '%s.%s'%(file_name.split('.')[0], file_type)
      os.rename(file_path, os.path.join('/'.join(file_path.split('/')[0:-1]), file_name))
      file_path = os.path.join('/'.join(file_path.split('/')[0:-1]), file_name)

    if 'jpg' in check_types and 'jpeg' not in check_types:
      check_types.append('jpeg')

    if file_type in check_types:
      return True, file_path

    return False, None
  elif check_types[0] in ['mp4', 'avi']:
    # VIDEO
    file_path = os.path.normpath(file_path)
    file_name = file_path.split('/')[-1]
    if file_name.split('.')[-1] == '':
      return False, None

    return True, file_path
  elif check_types[0] in ['txt']:
    # 文本
    file_path = os.path.normpath(file_path)
    file_name = file_path.split('/')[-1]
    if file_name.split('.')[-1] == '':
      return False, None

    return True, file_path

  else:
    # 不支持
    return False, None



# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : spider_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset.dataset import Dataset
import os
import sys
from antgo.utils.serialize import *
from PIL import Image
import imageio
import numpy as np
from antgo.ant.utils import *


class SpiderDataset(Dataset):
  def __init__(self, queue=None):
    super(SpiderDataset, self).__init__('test', '', None)
    self.queue = queue

  @property
  def size(self):
    return sys.maxsize

  def at(self, id):
    raise NotImplementedError

  def __spider_data_source(self, config):
    datasource_address = config['datasource_address']
    datasource_keywards = config['datasource_keywards']

    # 调用爬虫获取图像列表

    return []

  def data_pool(self):
    while True:
      try:
        data_pack = self.queue.get()
        data_source, config = data_pack
        data = []
        if data_source == 'spider':
          data = self.__spider_data_source(config)

        for image_file in data:
          image_data = Image.open(image_file)
          img_data = np.fromstring(image_data.tobytes(), dtype=np.uint8)
          img_data = img_data.reshape((image_data.size[1], image_data.size[0], len(image_data.getbands())))
          yield image_data, {}
      except:
        logger.error('fail receive data info')



# encoding=utf-8
# @Time    : 17-6-22
# @File    : __init__.py
# @Author  : jian<jian@mltalker.com>
import os
from pkgutil import walk_packages

__all__ = []


def global_import(name):
  p = __import__(name, globals(), locals(), level=1)
  lst = p.__all__ if '__all__' in dir(p) else dir(p)
  del globals()[name]
  for k in lst:
    globals()[k] = p.__dict__[k]

for _, module_name, _ in walk_packages(
        [os.path.dirname(__file__)]):
  if not module_name.startswith('_'):
    global_import(module_name)




# -*- coding: UTF-8 -*-
# @Time : 2018/4/28
# @File : queue_dataset.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.dataflow.dataset.dataset import Dataset
import os
import sys
from antgo.utils.serialize import *
from PIL import Image
import imageio
import numpy as np


class QueueDataset(Dataset):
  def __init__(self, queue=None):
    super(QueueDataset, self).__init__('test', '', None)
    self.queue = queue

  @property
  def size(self):
    return sys.maxsize

  def at(self, id):
    raise NotImplementedError

  def _video_iterator(self, video_path, request_param):
    reader = imageio.get_reader(video_path)
    for im in reader:
      img_data = np.fromstring(im.tobytes(), dtype=np.uint8)
      img_data = img_data.reshape((im.shape[0], im.shape[1], im.shape[2]))
      yield img_data if request_param is None else (img_data, request_param)

  def _parse_data(self, data, data_type, request_param):
    if data_type == 'VIDEO':
      return self._video_iterator(data, request_param)
    elif data_type == 'IMAGE':
      image_data = Image.open(data)
      img_data = np.fromstring(image_data.tobytes(), dtype=np.uint8)
      img_data = img_data.reshape((image_data.size[1], image_data.size[0], len(image_data.getbands())))
      return img_data if request_param is None else (img_data, request_param)
    elif data_type == 'FILE':
      with open(data, 'r') as fp:
        return fp.read() if request_param is None else (fp.read(), request_param)
    elif data_type == 'STRING':
      return data if request_param is None else (data, request_param)
    elif data_type == 'JSON':
      return data if request_param is None else (data, request_param)
    else:
      return None

  def data_pool(self):
    while True:
      data_pack = self.queue.get()
      if type(data_pack) == list and \
              (type(data_pack[0]) == list or type(data_pack[0]) == tuple):
        # multi-data
        storage = []
        for dd in data_pack:
          data, data_type, request_param = dd
          response_data = self._parse_data(data,data_type,request_param)
          storage.append(response_data)
        yield storage
      else:
        # single-data
        data, data_type, request_param = data_pack
        yield self._parse_data(data, data_type, request_param)


# -*- coding: UTF-8 -*-
# Time: 12/2/17
# File: tfdataset.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from . import Dataset
import tensorflow as tf
import tensorflow.contrib.slim as slim
import os
import sys


class TFRecordsReader(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(TFRecordsReader, self).__init__(train_or_test, dir, params)
    assert(train_or_test in ['train', 'val', 'test'])
    temp = getattr(self, 'tf_data_size', '700,700,3')
    self._data_size = [int(s) for s in temp.split(',')]
    self._data_type = tf.as_dtype(getattr(self, 'tf_data_type', 'uint8'))
    temp = getattr(self, 'tf_label_size', '700,700,1')
    self._label_size = [int(s) for s in temp.split(',')]
    self._label_type = tf.as_dtype(getattr(self, 'tf_label_type', 'uint8'))
    self._train_num = int(getattr(self, 'tf_train_num', 199600))
    self._val_num = int(getattr(self, 'tf_val_num', 10000))
    self._test_num = int(getattr(self, 'tf_test_num', 10000))
    self._pattern = getattr(self, 'tf_pattern', '*.tfrecord')
    self._has_format = bool(getattr(self, 'tf_format', False))
  
  @property
  def data_size(self):
    return self._data_size
  @data_size.setter
  def data_size(self, val):
    self._data_size = val
  
  @property
  def label_size(self):
    return self._label_size
  @label_size.setter
  def label_size(self, val):
    self._label_size = val
  
  @property
  def data_type(self):
    return self._data_type
  @data_type.setter
  def data_type(self, val):
    self._data_type = val
  
  @property
  def label_type(self):
    return self._label_type
  @label_type.setter
  def label_type(self, val):
    self._label_type = val
  
  @property
  def file_pattern(self):
    return self._pattern
  @file_pattern.setter
  def file_pattern(self,val):
    self._pattern = val
  
  @property
  def has_format(self):
    return self._has_format
  @has_format.setter
  def has_format(self, val):
    self._has_format = val
  
  def data_pool(self):
    raise NotImplementedError
  
  def at(self, id):
    raise NotImplementedError

  def split(self, split_params={}, split_method=""):
    raise NotImplementedError
  
  @property
  def size(self):
    if self.train_or_test == 'train':
      return self._train_num
    elif self.train_or_test == 'val':
      return self._val_num
    elif self.train_or_test == 'test':
      return self._test_num
    
    return self._train_num
    
  def model_fn(self, *args, **kwargs):
    # 1.step candidate data file list
    file_names = tf.train.match_filenames_once(os.path.join(self.dir, self.train_or_test, self.file_pattern))
    # dd= tf.get_default_session().run(file_names)
    
    # 2.step shuffle data file list
    filename_queue = tf.train.string_input_producer(file_names, shuffle= self.train_or_test == "train")

    # 3.step read from data file
    reader = tf.TFRecordReader()
    _, serialized_example = reader.read(filename_queue)
    
    # 4.step parse data
    if self.has_format:
      keys_to_features = {
        'image/encoded': tf.FixedLenFeature((), tf.string, default_value=''),
        'image/format': tf.FixedLenFeature((), tf.string, default_value='jpeg'),
        'label/encoded': tf.FixedLenFeature((), tf.string, default_value=''),
        'label/format': tf.FixedLenFeature((), tf.string, default_value='png'),
      }
      items_to_handlers = {
        'image': slim.tfexample_decoder.Image('image/encoded', 'image/format', channels=self.data_size[-1]),
        'label': slim.tfexample_decoder.Image('label/encoded', 'label/format', channels=self.label_size[-1])
      }
  
      decoder = slim.tfexample_decoder.TFExampleDecoder(keys_to_features, items_to_handlers)
      serialized_example = tf.reshape(serialized_example, shape=[])
      image, label = decoder.decode(serialized_example, ['image', 'label'])
      return image, label
    else:
      features = tf.parse_single_example(serialized_example,
                                         features={
                                            'image': tf.FixedLenFeature([], tf.string),
                                            'label': tf.FixedLenFeature([], tf.string),
                                         })
  
      image = tf.decode_raw(features['image'], self.data_type)
      if self.data_size is not None:
        image = tf.reshape(image, self.data_size)
  
      label = tf.decode_raw(features['label'], self.label_type)
      if self.label_size is not None:
        label = tf.reshape(label, self.label_size)
      
      return image, label


# -*- coding: UTF-8 -*-
# @Time    : 17-9-5
# @File    : generate.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant.base import *
from antgo.dataflow.basic import *
from antgo.utils.dht import *
from antgo.utils.serialize import *

if sys.version > '3':
    PY3 = True
else:
    PY3 = False

Config = config.AntConfig


class AntGenerate(AntBase):
  def __init__(self, ant_context,
               ant_name,
               ant_data_folder,
               dump_dir,
               ant_token,
               dataset_name):
    super(AntGenerate, self).__init__(ant_name, ant_context, ant_token)
    self.ant_data_folder = ant_data_folder
    self.ant_context.ant = self
    self.dataset_name = dataset_name
    self.dump_dir = dump_dir

  def start(self):
    # 1.step check basic parameters
    if self.context.data_generator is None:
      logger.error('must set data generator')
      return

    if self.dataset_name is None:
      logger.error('must set dsataset name')
      return

    # 2.step build experiment folder
    now_time_stamp = datetime.fromtimestamp(self.time_stamp).strftime('%Y%m%d.%H%M%S.%f')
    if not os.path.exists(os.path.join(self.dump_dir, now_time_stamp, self.dataset_name)):
      os.makedirs(os.path.join(self.dump_dir, now_time_stamp, self.dataset_name))
    experiment_folder = os.path.join(self.dump_dir, now_time_stamp, self.dataset_name)

    logger.info('dataset would be saved in %s'%now_time_stamp)

    # 3.step build dataset (rocksdb)
    # dataset "train", 'val', 'test', 'sample'
    for dataset_flag in ['train', 'val', 'test', 'sample']:
      dataset_generateor = self.context.data_generator(dataset_flag)
      if dataset_generateor is not None:
        logger.info('start build %s set for %s'%(dataset_flag, self.dataset_name))
        if not os.path.exists(os.path.join(experiment_folder, dataset_flag)):
          os.makedirs(os.path.join(experiment_folder, dataset_flag))

        if dataset_flag == 'sample':
          sample_set = []
          for data, label in dataset_generateor:
            sample_set.append((data, label))

          with open(os.path.join(experiment_folder, dataset_flag, 'sample'), 'wb') as fp:
            fp.write(dumps(sample_set))
        else:
          record_writer = RecordWriter(os.path.join(experiment_folder, dataset_flag))
          for data, label in dataset_generateor:
            record_writer.write(Sample(data=data, label=label))

        logger.info('finish build %s set for %s'%(dataset_flag, self.dataset_name))
      else:
        logger.warn('couldn build %s set for %s'%(dataset_flag, self.dataset_name))


    # dataset_url = ''
    # now_time_stamp = datetime.fromtimestamp(self.time_stamp).strftime('%Y%m%d.%H%M%S.%f')
    # if not os.path.exists(os.path.join(self.dump_dir, now_time_stamp)):
    #   os.makedirs(os.path.join(self.dump_dir, now_time_stamp))
    #
    # # preprocess data generator
    # categories = ['train', 'val', 'test', 'sample']
    # data_generators = {}
    # for category in categories:
    #   if category not in self.context.params.content:
    #     continue
    #
    #   data_generators[category] = {}
    #
    #   # data sample generator
    #   data_generators[category]['generator'] = self.context.data_generator(category)
    #   # data sample number
    #   data_generators[category]['num'] = self.context.params.content[category]['num']
    #   data_generators[category]['block'] = self.context.params.content[category]['block']
    #
    # # publish at dht
    # dataset_hash_code = dataset_upload_dht(self.dataset_name,
    #                                        data_generators,
    #                                        os.path.join(self.dump_dir,now_time_stamp))
    # dataset_url = 'ipfs://%s'%dataset_hash_code



# -*- coding: UTF-8 -*-
# @Time    : 2020-07-04 10:24
# @File    : batch_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import sys
import uuid
import json
import time
import math


class EntryApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    content = self.db['content']
    if content is None:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if 'dataset' not in content:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    dataset_and_results = content['dataset']
    dataset_names = list(dataset_and_results.keys())
    dataset_name = dataset_names[0]

    query_dataset_name = self.get_argument('dataset_name', None)
    if query_dataset_name is not None:
      dataset_name = query_dataset_name

    result_num = len(dataset_and_results[dataset_name])

    current_dataset_name = dataset_name

    if dataset_name not in self.db:
      self.db[dataset_name] = {}

    if 'num_in_page' not in self.db[dataset_name] or 'page_num' not in self.db[dataset_name]:
      max_num_in_page = self.db['max_num_in_page']
      max_page_num = self.db['max_page_num']

      page_num = int(math.ceil(result_num/max_num_in_page))
      if page_num > max_page_num:
        page_num = max_page_num

      num_in_page = int(math.ceil(result_num / page_num))
      self.db[dataset_name]['num_in_page'] = num_in_page
      self.db[dataset_name]['page_num'] = page_num

    num_in_page = self.db[dataset_name]['num_in_page']
    page_num = self.db[dataset_name]['page_num']
    page_index = 0
    start_index = 0 * num_in_page
    start_index = start_index if start_index < result_num else result_num-1
    end_index = 1*num_in_page
    end_inidex = end_index if end_index < result_num else result_num

    sample_num_in_dataset = {}
    for dataset_name in dataset_names:
      sample_num_in_dataset[dataset_name] = len(self.db['content']['dataset'][dataset_name])

    response_content = {
      'num_in_page': num_in_page,
      'page_num': page_num,
      'page_index': page_index,
      'dataset_name': dataset_names,
      'sample_num_in_dataset': sample_num_in_dataset,
      'current_dataset_name': current_dataset_name,
      'result_num': result_num,
      'result_list': dataset_and_results[dataset_name][start_index: end_index]
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class PageApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    page_index = self.get_argument('page_index', None)
    page_index = int(page_index)
    dataet_name = self.get_argument('dataset_name', None)
    if page_index is None or dataet_name is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    content = self.db['content']
    if content is None:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if dataet_name not in self.db:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if 'num_in_page' not in self.db[dataet_name] or 'page_num' not in self.db[dataet_name]:
      self.response(RESPONSE_STATUS_CODE.RESOURCE_NOT_FOUND)
      return

    if page_index >= self.db[dataet_name]['page_num']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    result_num = len(self.db['content']['dataset'][dataet_name])
    start_index = page_index*self.db[dataet_name]['num_in_page']
    start_index = start_index if start_index < result_num else result_num - 1
    end_index = (page_index+1) * self.db[dataet_name]['num_in_page']
    end_inidex = end_index if end_index < result_num else result_num

    dataset_names = list(content['dataset'].keys())
    sample_num_in_dataset = {}
    for dataset_name in dataset_names:
      sample_num_in_dataset[dataset_name] = len(self.db['content']['dataset'][dataset_name])

    response_content = {
      'num_in_page': self.db[dataet_name]['num_in_page'],
      'page_num': self.db[dataet_name]['page_num'],
      'page_index': page_index,
      'dataset_name': dataset_names,
      'current_dataset_name': dataet_name,
      'sample_num_in_dataset': sample_num_in_dataset,
      'result_num': result_num,
      'result_list': content['dataset'][dataet_name][start_index: end_index]
    }
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=response_content)


class ConfigApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    config_data = self.get_argument('config_data', None)
    if config_data is None:
      self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
      return

    config_data = json.loads(config_data)

    self.db['content'] = config_data
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class PingApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def batch_server_start(batch_dump_dir, server_port):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    batch_static_dir = os.path.join(batch_dump_dir, 'batch')

    # 2.step launch web server
    db = {'content': {}, 'max_page_num': 30, 'max_num_in_page': 50}
    settings = {
      'static_path': os.path.join(batch_static_dir, 'static'),
      'port': server_port,
      'cookie_secret': str(uuid.uuid4()),
      'db': db
    }

    app = tornado.web.Application(handlers=[(r"/batch-api/entry/", EntryApiHandler),
                                            (r"/batch-api/dataset/", EntryApiHandler),
                                            (r"/batch-api/page/", PageApiHandler),
                                            (r"/batch-api/config/", ConfigApiHandler),
                                            (r"/batch-api/ping/", PingApiHandler),
                                            (r'/(.*)', tornado.web.StaticFileHandler,
                                             {"path": batch_static_dir, "default_filename": "index.html"}),],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('demo is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass